/*
	neo.c	Copyright (C) 2002 By Gregory Dietsche <email: gforce@calc.org>
					Webpage: http://gforce.calc.org/

DESCRIPTION:
	This program is very powerful. It can apply a patch at the beginning of
	any program on a TI-89, TI-92PLUS, or TI-V200. Once installed, this patch,
	is executed immediately before the patched program is run.
	
	You can thank my development of KerNO for this program. The original program (patch.c)
	was designed to allow kerno to run ppg (compressed) programs. This idea was, however,
	dropped for lack of good reasons. (Seriously) (read that sentance carefully)
	(it does not mean what it says, but it says what it means)

HOW TO USE:
	type: neo("prog") where prog is the name of the ASM program to be patched.

LIMITATIONS:
	The patch must be pc-relative and contain no relocation table.

COMPILING:
	This program was compiled using TIGCC Version 0.94 Beta 18
	GCC Switches: -W -Wall -Os -mpcrel -fomit-frame-pointer -DUSE_ENGLISH

LEGAL INFORMATION:
	The author makes no representations or warranties about the suitability
	of the software, either express or implied. The author shall not be liable 
	for any damages suffered as a result of using or distributing this software.

	This program and or source may not be modified and or distributed unless permission
	is granted by the author of this program. This is intended to prevent misuse
	of the program and included source code.

	If distributed, this program must not be sold as or a part of a sold package.

	This program and or modified versions may be distributed only when the
	above terms and conditions have been fully met. 


*/

/*
	This souce code is a bit of a mess. Sorry.
	I have the structures in place to implement a full scale patching program
	which could contain multiple patches, and be able to install/uninstall the patches
	independantly. Also, I'm sure that there is a large amount of size optimiztaion that
	could be done.
*/
#ifdef USE_ENGLISH
	#define STR_PATCH_VER		"Neo v1.0a by Greg Dietsche"
	#define STR_PATCHED			"Patch TI-Bullet Time Installed"
	#define STR_UNPATCHED		"Patch TI-Bullet Time Removed"
	#define STR_LOW_RAM 		"Low RAM"
	#define STR_ARCHIVED		"Archived Var"
	#define STR_ODD_SIZE		"Var Length Must Be Odd"
	#define STR_ARG_ERROR 	"patch(\"progname\")"
	#define STR_BAD_FILE		"Bad Var Type"
	#define STR_NOT_FOUND 	"Bad Var Name"
#endif

#define USE_BULLET_TIME_PATCH
//#define USE_KERNO_PATCH

#define OPTIMIZE_ROM_CALLS
#define NO_EXIT_SUPPORT
#define NO_CALC_DETECT
#define MIN_AMS 200			//require ams 2.xx !! this is because I used DlgMessage with long strings

#define USE_TI89
#define USE_TI92PLUS
#define USE_V200

#define fatal(s) {DisplayText(s); return;}
#define PATCH_SIZE (PatchEnd-PatchStart)

#include <tigcclib.h>         // Include All Header Files

typedef struct {
	unsigned short programsize;		/* The size of the program to be patched*/
	const short bra;							/* The bra to patch code */
	unsigned long sig;						/* Magic Number Signature used to Identify Patched Files */
  short size;										/* Size of the installed patch */
	char str[];										/* ASCIZ string */
}PATCH_DAT;

extern void PatchStart(void);
extern void PatchEnd(void);
extern char PatchStr[];
void InstallPatch(volatile HANDLE h);
void UninstallPatch(volatile HANDLE h);

/* This routine requires that the program size in the var-link be odd! */
inline static void FixRelocTable(short delta_offset asm("%d0"), void * tag_ptr asm("%a0"))
{
	while(*--(unsigned short*)tag_ptr)
	{
		*(unsigned short*)tag_ptr+=delta_offset;
		*--(unsigned short*)tag_ptr+=delta_offset;
	}
}

static void DisplayText(const char * const str asm("%a0")){ DlgMessage(STR_PATCH_VER, str, BT_OK, 0); }

// Main Function
void _main(void)
{

	char tmpstr[300], *fname, *src;
	volatile HANDLE h;
	ESI argptr;
	SYM_ENTRY *SymPtr;

	InitArgPtr(argptr);
	
	if (GetArgType(argptr) != STR_TAG)
		fatal(STR_ARG_ERROR);

	fname=(char*)GetStrnArg(top_estack);

//ignore anything after the last (
	if((src = strchr(fname,'('))) *src = 0;//using src out of context...

//convert to TIOS sytle file name
	tmpstr[0] = 0;
	strcpy(tmpstr+1, fname);
	
//Find the program in the vat...
	if(!(SymPtr = DerefSym(SymFind(tmpstr+strlen(fname)+1)))) fatal(STR_NOT_FOUND);
	if(*(unsigned char*)(src = HToESI(h = SymPtr->handle))!=ASM_TAG) fatal(STR_BAD_FILE);
	if(SymPtr->flags.bits.archived)fatal(STR_ARCHIVED);

	PATCH_DAT *p=HeapDeref(h);
	if(p->size==PATCH_SIZE && p->sig==0 && !strcmp(p->str, PatchStr))	//is it installed?
		UninstallPatch(h);
	else
		InstallPatch(h);
}

void InstallPatch(volatile HANDLE h)
{
	unsigned short length;
	unsigned char *src;
	
//Prepare to patch		
	src=HeapDeref(h);
	length=*(unsigned short*)src;
	
	if(!(length%2)) fatal(STR_ODD_SIZE);//for fixreloctable func
	
	if((HeapRealloc(h,length+(2+PATCH_SIZE)))==H_NULL) fatal(STR_LOW_RAM);
	
  src=HeapDeref(h);
  *(unsigned short*)src+=PATCH_SIZE;
  
//install the patch
	memmove(src+(2+PATCH_SIZE),src+2,length);
	memcpy(src+2,PatchStart,PATCH_SIZE);

//Update Relocation Table to reflect the new offsets
	FixRelocTable(PATCH_SIZE,(unsigned char*)src+*(unsigned short*)src+1);	//patch_size, baseptr + length + 1	
	
	fatal(STR_PATCHED);
}

void UninstallPatch(volatile HANDLE h)
{
	unsigned short length;
	unsigned char *src;
	
	src=HeapDeref(h);
	
  *(unsigned short*)src-=PATCH_SIZE;
  length=*(unsigned short*)src;

//remove the patch
	memmove(src+2,src+(2+PATCH_SIZE),length);

//Update Relocation Table to reflect the new offsets
	FixRelocTable(-PATCH_SIZE,(unsigned char*)src+(length+1));	//patch_size, baseptr + length + 1	

	HeapRealloc(h,length+2);
	
	fatal(STR_UNPATCHED);
}

#ifdef USE_BULLET_TIME_PATCH
asm("
/* Patch a game with the trace interrupt */
/* and make it run very slowly :) */
PatchStart:

 bra.s _PatchStart
.long 0
.word PatchEnd-PatchStart
PatchStr:
.asciz \"TI-Bullet Time\"
InfoStr:
.asciz \"TI-Bullet Time v1.0 By Greg Dietsche\"
.EVEN

_PatchStart:
 lea.l OldTrace(%pc),%a0
 move.l 0x40024,(%a0)+	/* Trace Interrupt 	*/
 move.l 0x40084,(%a0)		/* Trap #1					*/
 
 lea.l NewTrapOne(%pc),%a0
 move.l %a0,0x40084

 lea NewTrace(%pc),%a0	/* this is necessary; Games Like Phoenix Crash Without it 			*/
 move.l %a0,0x40024			/* otherwise, this could be omitted to save a few bytes 				*/
 												/* I haven't tried it, but i suspect that commenting out these 	*/
 												/* lines may allow this program to work with DoorsOS Programs 	*/
 
 /*trap #12					uncomment this for immediate trace instruction execution
 move #0x8000,%sr		games like Phoenix crash with this uncommented. */
 
 move.l #0x00010000,-(%sp)
 pea InfoStr(%pc)
 pea PatchStr(%pc)
 move.l 0xc8,%a0
 move.l 0x1b4*4(%a0),%a0
 jsr (%a0)
 lea.l 12(%sp),%sp
 
 bsr PatchEnd
 nop							/* fix for programs that use the tigcc return value hack (they won't crash anymore) */
 nop
 nop
 trap #12
 and #0x5FFF,%sr	/* unset supervisor and trace bits */
 
 lea.l OldTrace(%pc),%a0
 move.l (%a0)+,0x40024
 move.l (%a0),0x40084
 
 rts
 
NewTrapOne:
 move.l %a0,-(%sp)
 lea NewTrace(%pc),%a0
 move.l %a0,0x40024			/* reassert that i have control of the trace vector */
 move.l (%sp)+,%a0
 
 bset #15,(%sp)					/* force rte to return with supervisor mode set */
 
 move.l OldTrapOne(%pc),-(%sp)
 rts
 
NewTrace:	/* 38 clocks */
 rte		  /* 24 clocks */
 /* 62 clock cycles wasted between _every_ user mode instruction */

OldTrace:
.long 0
OldTrapOne:
.long 0

PatchEnd:
");
#endif

#ifdef USE_KERNO_PATCH
//an old, bad idea... but it has given light to a new idea... stay tuned
asm("
/*Patch is 26 bytes long */

PatchStart:
 bsr.s LoaderCode			/* 2 bytes */
 .ascii \"68kN\"			/* KerNO Signature 0x36386B4E = 68kN */
 .word 1							/* Flags Field */

LoaderCode:
 cmp.w #0x4B4E,0x32		/* Check for KN signature */
 beq.s CallKerNO

 add.l	#4,%a7				/*	KerNO was not found, so restore the stack, and jump to the main program */
 bra.s PatchEnd
 
CallKerNO:
 move.l 0x34,%a0				/* KerNO exec pointer (when using vector 0x34, no version check is necessary */
 jmp (%a0)

PatchEnd:

");
#endif
